package com.yuandian.dataflow.statemachine;

import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Function;

import com.alipay.remoting.exception.CodecException;
import com.alipay.remoting.serialization.SerializerManager;
import com.alipay.sofa.jraft.Closure;
import com.alipay.sofa.jraft.Iterator;
import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.conf.Configuration;
import com.alipay.sofa.jraft.core.StateMachineAdapter;
import com.alipay.sofa.jraft.entity.LeaderChangeContext;
import com.alipay.sofa.jraft.error.RaftException;
import com.alipay.sofa.jraft.error.RemotingException;
import com.alipay.sofa.jraft.storage.snapshot.SnapshotReader;
import com.alipay.sofa.jraft.storage.snapshot.SnapshotWriter;
import com.yuandian.dataflow.statemachine.rpc.RaftResponse;
import com.yuandian.dataflow.statemachine.rpc.SyncWorkerStateProcessor.RequestWorkerState;
import com.yuandian.dataflow.statemachine.rpc.SyncStateProcessor.RequestState;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.statemachine.state.StateFactory;
import com.yuandian.dataflow.statemachine.state.WorkerState;

import lombok.var;
import lombok.extern.slf4j.Slf4j;

/**
 * Counter state machine.
 *
 * @author boyan (boyan@alibaba-inc.com)
 *
 *         2018-Apr-09 4:52:31 PM
 */
@Slf4j
public class StateMachine extends StateMachineAdapter {

    // private static final Logger LOG = LoggerFactory.getLogger(StateMachine.class);

    /**
     * State value 全局使用的唯一状态
     */
    private State state = new State();
    /**
     * Leader term
     */
    private final AtomicLong leaderTerm = new AtomicLong(-1);

    public boolean isLeader() {
        return this.leaderTerm.get() > 0;
    }

    /**
     * Returns current value. 只有Get 操作状态由协议流程决定 Apply
     */
    public State getState() {
        return state;
    }

 
 
 
    @Override
    @SuppressWarnings("unchecked")
    public void onApply(final Iterator iter) {
        while (iter.hasNext()) {

            if (iter.done() != null) {
                // This task is applied by this node, get value from closure to avoid additional
                // parsing.
                var closure = (SyncClosure<State>)iter.done(); // 只支持单一个State. 全状态机只支持一种提交
                // log.info("SyncDataClosure(done) taskState:{} leaderTerm:{} {}",this.state, this.leaderTerm, closure); 
                this.state = closure.getValue();
                closure.success(state);
                closure.run(Status.OK());
            } else {
                 // Have to parse FetchAddRequest from this user log.
                final ByteBuffer data = iter.getData();
                try {
        
                state = SerializerManager.getSerializer(SerializerManager.Hessian2).deserialize(
                data.array(), State.class.getName());
                // log.info("SyncDataClosure(null) taskState:{} leaderTerm:{}", state, this.leaderTerm);
               
                
                } catch (CodecException e) {
                    e.printStackTrace();
                }
               
            }

            iter.next();
        }
    }

    @Override
    public void onSnapshotSave(final SnapshotWriter writer, final Closure done) {
        return;
    }

    @Override
    public void onError(final RaftException e) {
        log.debug("Raft error: {}", e, e);
    }

    @Override
    public boolean onSnapshotLoad(final SnapshotReader reader) {
        return true;
    }

    @Override
    public void onLeaderStart(final long term) {
        log.debug("onLeaderStart {}", StateServerFactory.getCluster().getServerId());
        this.leaderTerm.set(term);

        if(StateFactory.getMasterExecute().isAlive()) {  
            StateFactory.getMasterExecute().interrupt();
        }
 
        var ss =  StateServerFactory.getStateServer();
        ss.useFsmStateNotLock((state)->{
            var ws = state.getWorkers().get( ss.getCluster().getServerId() );
            if(ws == null) {
                ws = new WorkerState(ss.getCluster().getServerId());
                state.getWorkers().put(ss.getCluster().getServerId(), ws);
            }

            ss.applyState(state, new SyncClosure<State>() {
                @Override
                public void run(Status status) {
                    log.debug("master update workerstate: {}", status);
                }

            });

        });

        StateFactory.getMasterExecute().start();
 
        super.onLeaderStart(term);
    }

    @Override
    public void onLeaderStop(final Status status) {
        log.debug("onLeaderStop {}", StateServerFactory.getCluster().getServerId());
        this.leaderTerm.set(-1);
        super.onLeaderStop(status);

        if(StateFactory.getMasterExecute().isAlive()) {
            StateFactory.getMasterExecute().interrupt();
        }

  
    }

    

    @Override
    public void onShutdown() {
        log.debug("onShutdown");
        super.onShutdown();
    }
 

    @Override
    public void onStartFollowing(LeaderChangeContext ctx) {
        log.debug("[onStartFollowing] {} {}", ctx, StateServerFactory.getCluster().getServerId());
        try {

            if(StateFactory.getMasterExecute().isAlive()) {
                StateFactory.getMasterExecute().interrupt();
            }

            var ss = StateServerFactory.getStateServer();
            var ws = new WorkerState(ss.getCluster().getServerId());
            var request = new RequestWorkerState();
            request.setWorkerState(ws);
            log.debug("my: {} leader id {}",ss.getCluster().getServerId(), ss.getNode().getLeaderId());
            RaftResponse resp;
            resp = (RaftResponse)StateServerFactory.getStateServer().getRpcClient().invokeSync(ss.getNode().getLeaderId().getEndpoint(), request,  5000);
            if(resp == null) {
                log.debug("{} set WorkerState is error", resp);
            }
            log.debug("WorkerState is {}", resp);
            return;
        } catch (InterruptedException | RemotingException e) {
            e.printStackTrace();
        }

        super.onStartFollowing(ctx);
    }

    @Override
    public void onConfigurationCommitted(Configuration conf) {
        super.onConfigurationCommitted(conf);
    }

    @Override
    public void onStopFollowing(LeaderChangeContext ctx) {
        log.debug("{} {}",ctx, StateServerFactory.getCluster().getServerId());
        super.onStopFollowing(ctx);
    }

 

   



}
